"""
Schemas related to model endpoints
"""
import typing as t

from pydantic import UUID4, BaseModel

from taktile_types.enums.endpoint import (
    ArrowFormatKinds,
    EndpointKinds,
    ProfileKinds,
)


class EndpointInfoSchema(BaseModel):
    """EndpointInfoSchema.
    Endpoint type of /info endpoint

    Attributes
    ----------

    name : str
        name of endpoint
    position : t.Optional[int]
        position of endpoint relative to others
    path : str
        path of endpoint
    kind : EndpointKinds
       kind of endpoint
    profile_kind : t.Optional[ProfileKinds]
       profile type of this endpoint
    explain_path : t.Optional[str]
        where is the explainer endpoint
    response_kind: t.Optional[ArrowFormatKinds]
        what type of arrow response does this endpoint return
    input_names: t.List[str] = []
        list of input names
    output_names: t.Optional[str] = None
        output name
    profile_columns : t.Optional[t.List[str]] = None
        columns that have been profiled
    input_example : t.Optional[t.List[t.Dict[str, t.Any]]] = None
        json input example for this endpoint
    output_example : t.Optional[t.List[t.Any]] = None
        example output of endpoint
    explainer_example : t.Optional[t.List[t.Any]] = None
        example input for explainer
    tags : t.Optional[t.List[str]]
        list of tags of this endpoint
    """

    name: str
    position: t.Optional[int]
    path: str
    kind: EndpointKinds
    profile_kind: t.Optional[ProfileKinds]
    explain_path: t.Optional[str]
    response_kind: t.Optional[ArrowFormatKinds]
    input_names: t.List[str] = []
    output_names: t.Optional[str] = None
    profile_columns: t.Optional[t.List[str]] = None
    input_example: t.Optional[t.List[t.Dict[str, t.Any]]] = None
    output_example: t.Optional[t.List[t.Any]] = None
    explainer_example: t.Optional[t.List[t.Any]] = None
    tags: t.Optional[t.List[str]]


class InfoEndpointResponseModel(BaseModel):
    """InfoEndpointResponseModel.
    Return type of /info endpoint

    Attributes
    ----------
    schema_version:: str
        response version of this response
    taktile_cli : str
        taktile_cli version installed
    profiling : str
        profiling version installed
    endpoints : t.List[EndpointInfoSchema]
        list of endpoints on this model
    """

    schema_version: str
    taktile_cli: str
    profiling: str
    endpoints: t.List[EndpointInfoSchema]


class EndpointDeploymentApi(BaseModel):
    """EndpointDeploymentApi.
    Endpoint type in deployment-api

    Attributes
    ----------
    id : UUID4
        unique identifier of endpoint
    name : str
        name of endpoint
    git_hash : str
        sha of git commit of this endpoint
    """

    id: UUID4
    name: str
    git_hash: str


class ModelEndpoint(BaseModel):
    """ModelEndpoint.
    Extended endpoint type on deployment-api

    Attributes
    ----------
    kind : t.Optional[EndpointKinds]
        endpoint kind
    position : int
        order of endpoints
    tags : t.List[str]
        tags of an endpoint (see OAS)
    """

    kind: t.Optional[EndpointKinds]
    position: int
    tags: t.List[str]



"""Datastructures and validatation for taktile config"""
import typing as t

from pydantic import BaseModel
from pydantic.class_validators import root_validator

from taktile_types.schemas.repository.resources import (
    SIZE_MAP,
    ResourceRequirement,
)


class ServiceWithSizing(BaseModel):
    """Service scaling and sizing information"""

    replicas: int
    requests: ResourceRequirement

    @root_validator(pre=True)
    def service_validator(
        cls, vals
    ):  # pylint: disable=no-self-argument,no-self-use
        """Validates individual service"""
        instance_type_set = "instance_type" in vals
        requests_set = False
        if "requests" in vals:
            requests = vals["requests"]
            if isinstance(requests, dict):
                requests_set = (
                    requests["cpu"] is not None
                    and requests["memory"] is not None
                )
            elif isinstance(requests, ResourceRequirement):
                requests_set = (
                    requests.cpu is not None and requests.memory is not None
                )
        if instance_type_set and requests_set:
            raise ValueError(
                "Both instance type and requests were found"
                " during validation of service config"
            )
        if not instance_type_set and not requests_set:
            raise ValueError(
                "Neither instance type nor requests were found"
                " during validation of service config"
            )
        if instance_type_set:
            instance_type = vals["instance_type"].split(".")
            if len(instance_type) < 2:
                raise ValueError("Instance type must contain a dot")
            if len(instance_type) > 2:
                raise ValueError(
                    "Instance type cannot contain more than one dot"
                )
            instance_class, instance_size = instance_type
            try:
                vals["requests"] = SIZE_MAP[instance_class][instance_size]
            except KeyError:
                raise ValueError(  # pylint: disable=raise-missing-from
                    "Invalid instance type"
                )
        if vals["replicas"] < 0:
            raise ValueError("Replica count must be non-negative")
        return vals


class RestServiceConfig(ServiceWithSizing):
    """Additional info for REST services"""

    max_replicas: int


class TktlServiceConfig(BaseModel):
    """Config info for all services"""

    rest: RestServiceConfig
    arrow: ServiceWithSizing


class TaktileYAMLConfig(BaseModel):
    """data structure for full taktile config yaml"""

    deployment_prefix: t.Optional[str]
    undeployment_prefix: t.Optional[str]
    service: TktlServiceConfig
    version: str



"""
Schemas related to the /models endpoint
"""
import typing as t
from datetime import datetime

from pydantic import UUID4, BaseModel

from taktile_types.enums.endpoint import EndpointKinds
from taktile_types.enums.repository.access import AccessKind
from taktile_types.schemas.deployment.endpoint import EndpointDeploymentApi
from taktile_types.schemas.deployment.test import TestResultResponse
from taktile_types.schemas.repository.pull_request import PullRequest


class ModelEndpoint(EndpointDeploymentApi):
    """ModelEndpoint.
    Extended endpoint model

    Attributes
    ----------
    kind : t.Optional[EndpointKinds]
        The kind of this endpoint
    position : int
        Its position relative to the other endpoints
    tags : t.List[str]
        A list of tags of this endpoint
    """

    kind: t.Optional[EndpointKinds]
    position: int
    tags: t.List[str]


class ModelDeployment(BaseModel):
    """ModelDeployment.
    A deployed model

    Attributes
    ----------
    id : UUID4
        unique identifier
    created_at : datetime
        when was this created
    status : t.Optional[str]
       status of this deployment
    public_docs_url : t.Optional[str]
       url to the public documentation
    rest_cpu_request : t.Optional[str]
       cpu requested
    rest_memory_request : t.Optional[str]
       memory requested
    rest_gpu_request : t.Optional[str]
        gpu requested
    arrow_cpu_request : t.Optional[str]
       cpu requested
    arrow_memory_request : t.Optional[str]
       memory requested
    arrow_gpu_request : t.Optional[str]
       gru requested
    rest_replicas : t.Optional[int]
       rest default replicas requested
    max_rest_replicas : t.Optional[int]
       rest max replicas
    arrow_replicas : t.Optional[int]
       arrow replicas
    git_ref : str
       git ref of this deployment
    check_run_reference_id : t.Optional[int]
       GH id of latest check run
    commit_hash : str
       commit hash of this deployment
    endpoints : t.List[ModelEndpoint]
       endpoints of this deployment
    tests : t.Optional[TestResultResponse]
       tests of this deployment
    running_commit : t.Optional[str]
       currently running commit
    rest_runtime_logs_link : t.Optional[str]
       link to rest runtime logs on Grafana
    grpc_runtime_logs_link : t.Optional[str]
       link to arrow runtime logs on Grafana
    build_logs_link : t.Optional[str]
       link to build logs on Grafana
    rest_runtime_dashboard_link : t.Optional[str]
       link to rest dashboard on grafana
    grpc_runtime_dashboard_link : t.Optional[str]
       link to grpc dashboard on grafana
    """

    id: UUID4
    created_at: datetime
    status: t.Optional[str]
    public_docs_url: t.Optional[str]
    rest_cpu_request: t.Optional[str]
    rest_memory_request: t.Optional[str]
    rest_gpu_request: t.Optional[str]
    arrow_cpu_request: t.Optional[str]
    arrow_memory_request: t.Optional[str]
    arrow_gpu_request: t.Optional[str]
    rest_replicas: t.Optional[int]
    max_rest_replicas: t.Optional[int]
    arrow_replicas: t.Optional[int]
    git_ref: str
    check_run_reference_id: t.Optional[int]
    commit_hash: str
    endpoints: t.List[ModelEndpoint]
    tests: t.Optional[TestResultResponse]
    running_commit: t.Optional[str]

    rest_runtime_logs_link: t.Optional[str]
    grpc_runtime_logs_link: t.Optional[str]
    build_logs_link: t.Optional[str]
    rest_runtime_dashboard_link: t.Optional[str]
    grpc_runtime_dashboard_link: t.Optional[str]


class Model(BaseModel):
    """Model.
    Response of the model endpoint

    Attributes
    ----------
    id : UUID4
       unique identifier
    ref_id : int
       GH identifier of model
    repository_name : str
       name of repository
    repository_owner : str
        repository owner of model
    repository_default_branch : t.Optional[str] = None
        default branch of model
    repository_description : t.Optional[str] = None
        GH description of repository
    access : AccessKind
        Access type the requesting user has to this model
    pull_requests : t.List[PullRequest]
        A list of pull requests of this model
    deployments: t.List[ModelDeployment]
        A list of deployments of this model
    """

    id: UUID4
    ref_id: int
    repository_name: str
    repository_owner: str
    repository_default_branch: t.Optional[str] = None
    repository_description: t.Optional[str] = None
    access: AccessKind
    pull_requests: t.List[PullRequest]
    deployments: t.List[ModelDeployment]



"""Pull request related datastructures"""
from pydantic import UUID4, BaseModel


class PullRequest(BaseModel):
    """PullRequest.

    A data structure to hold a pull request
    """

    github_id: int
    url: str
    title: str
    github_number: int
    source_branch: str
    target_branch: str

    repository_id: UUID4



"""Repository related datastructures"""
import typing as t

from pydantic import UUID4, BaseModel

from taktile_types.enums.repository.access import AccessKind
from taktile_types.schemas.repository.pull_request import PullRequest


class Repository(BaseModel):
    """A repository"""

    id: UUID4
    name: str
    full_name: str
    repository_id: int
    private: bool
    installation_id: UUID4
    default_branch: t.Optional[str] = None
    description: t.Optional[str] = None

    def get_principal_and_name(self) -> t.Tuple[str, str]:
        """get_principal_and_name.
        Retrieve principal and name of a repository. E.g.
        "taktile-org/sample-repo" -> "taktile-org", "sample-repo"

        Returns
        -------
        t.Tuple[str, str] - principal, name

        """
        principal_and_name = self.full_name.split("/")
        if len(principal_and_name) != 2:
            raise ValueError(
                f"Couldn't figure out principal and name from {self.full_name}"
            )
        return principal_and_name[0], principal_and_name[1]


class RepositoryExtended(Repository):
    """Repository with access and PR information"""

    access: AccessKind
    pull_requests: t.List[PullRequest]



"""
Enums related to deployment-api tests
"""
from taktile_types.enums.common import ExtendedEnum


class TestOutcomeColor(ExtendedEnum):
    """TestOutcomeColor.
    Color coded test outcome
    """

    GREEN = "green"
    RED = "red"
    YELLOW = "yellow"


class TestTypes(ExtendedEnum):
    """TestTypes.
    Type of test
    """

    INTEGRATION = "integration"
    UNIT = "unit"
